/* GPS 0.1 - Gravitational Particle Simulator
 * Copyright (C) 2000 by Fabio Mancinelli <mancinel@univaq.it>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef HAVE_CONFIG_H
#include<config.h>
#endif

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<gps.h>
#include<rk.h>


/******************************************************************************
  Static function prototypes
 ******************************************************************************/

static void gps_free_particle(gpointer data, gpointer user_data);
static void gps_write_particle(gpointer data, gpointer user_data);

/******************************************************************************
  Exported functions
 ******************************************************************************/

GPS *gps_new(void) {
        GPS *gps;
        
        /* Allocate the memory for the structure */
        gps = (GPS *)g_malloc(sizeof(GPS));
        
        /* Initialize the structure with default values */
        gps->g = 1.0;
        gps->step_size = 0.01;
        gps->variable_step_size = TRUE;
        gps->max_step_size = 0.01;
        gps->separation = 0.1;
        gps->particles_list = NULL;
        
        vector_set(&gps->camera, 5.0, 5.0, 5.0);
        vector_set(&gps->target, 0.0, 0.0, 0.0);
        gps->fov = 80.0;
        
        return gps;
};

void gps_free(GPS *gps) {
        g_list_foreach(gps->particles_list, gps_free_particle, NULL);
        g_free(gps);
};

gint gps_load(gchar *file_name, GPS *gps) {
        FILE *input;
        static gchar buffer[128];
        int particles;
        PARTICLE *particle;
        int i;
        int j;
        
        /* Open the file */
        input = fopen(file_name, "rt");
        if(input == NULL) return FALSE;
        
        /* Read the signature */
        fgets(buffer, 128, input);
        if(strncmp("GPS", buffer, 3) != 0) {
                fclose(input);
                return FALSE;
        };
        
        /* Read GPS parameters */
        fgets(buffer, 128, input);
        gps->g = atof(buffer);
        fgets(buffer, 128, input);
        gps->step_size = atof(buffer);
        fgets(buffer, 128, input);
        gps->variable_step_size = atoi(buffer);
        fgets(buffer, 128, input);
        gps->max_step_size = atof(buffer);
        fgets(buffer, 128, input);
        gps->separation = atof(buffer);

        fgets(buffer, 128, input);
        gps->camera.x = atof(buffer);
        fgets(buffer, 128, input);
        gps->camera.y = atof(buffer);
        fgets(buffer, 128, input);
        gps->camera.z = atof(buffer);
        
        fgets(buffer, 128, input);
        gps->target.x = atof(buffer);
        fgets(buffer, 128, input);
        gps->target.y = atof(buffer);
        fgets(buffer, 128, input);
        gps->target.z = atof(buffer);
        
        fgets(buffer, 128, input);
        gps->fov = atof(buffer);
        
        /* Get the number of the particles in the system */
        fgets(buffer, 128, input);
        particles = atoi(buffer);
        
        /* Read all the particle data */
        for(i = 0; i < particles; i++) {
                particle = (PARTICLE *)g_malloc(sizeof(PARTICLE));
                fgets(buffer, 128, input);
                buffer[strlen(buffer) - 1] = '\0';
                strcpy(particle->name, buffer);
                fgets(buffer, 128, input);
                particle->mass = atof(buffer);
                fgets(buffer, 128, input);
                particle->color[0] = atof(buffer);
                fgets(buffer, 128, input);
                particle->color[1] = atof(buffer);
                fgets(buffer, 128, input);
                particle->color[2] = atof(buffer);
                fgets(buffer, 128, input);
                particle->position.x = atof(buffer);
                fgets(buffer, 128, input);
                particle->position.y = atof(buffer);
                fgets(buffer, 128, input);
                particle->position.z = atof(buffer);
                fgets(buffer, 128, input);
                particle->velocity.x = atof(buffer);
                fgets(buffer, 128, input);
                particle->velocity.y = atof(buffer);
                fgets(buffer, 128, input);
                particle->velocity.z = atof(buffer);
                
                /* Initialize particle's trail */
                for(j = 0; j < TRAIL_MAX_POINTS; j++) {
                        vector_copy(&particle->trail[j], &particle->position);
                };
                
                gps_add_particle(particle, gps);
        };
        
        fclose(input);
        
        return TRUE;
};

gint gps_save(gchar *file_name, GPS *gps) {
        FILE *output;
        
        output = fopen(file_name, "wt");
        if(output == NULL) return FALSE;
        
        fprintf(output, "GPS\n");
        
        fprintf(output, "%f\n", gps->g);
        fprintf(output, "%f\n", gps->step_size);
        fprintf(output, "%i\n", gps->variable_step_size);
        fprintf(output, "%f\n", gps->max_step_size);
        fprintf(output, "%f\n", gps->separation);
        
        fprintf(output, "%f\n", gps->camera.x);
        fprintf(output, "%f\n", gps->camera.y);
        fprintf(output, "%f\n", gps->camera.z);
        
        fprintf(output, "%f\n", gps->target.x);
        fprintf(output, "%f\n", gps->target.y);
        fprintf(output, "%f\n", gps->target.z);
        
        fprintf(output, "%f\n", gps->fov);
                        
        fprintf(output, "%i\n", g_list_length(gps->particles_list));
        
        g_list_foreach(gps->particles_list, gps_write_particle, output);
        
        fclose(output);
        
        return TRUE;
};

void gps_add_particle(PARTICLE *particle, GPS *gps) {
        gps->particles_list = g_list_append(gps->particles_list, particle);
};

void gps_delete_particle(PARTICLE *particle, GPS *gps) {
        gps->particles_list = g_list_remove(gps->particles_list, particle);
        g_free(particle);
};

void gps_update(GPS *gps) {
        PARTICLE *particle;
        int i;
        int j;
        
        /* Update particle trails */
        for(i = 0; i < g_list_length(gps->particles_list); i++) {
                particle = (PARTICLE *)g_list_nth_data(gps->particles_list, i);
                
                for(j = (TRAIL_MAX_POINTS - 2); j >= 0; j--) {
                        vector_copy(&particle->trail[j + 1], &particle->trail[j]);
                };
                vector_copy(&particle->trail[0], &particle->position);
        };
                
        /* Update particle positions */
        rk_step(gps);
};        

        

        

/******************************************************************************
  Static functions
 ******************************************************************************/

static void gps_free_particle(gpointer data, gpointer user_data) {
        PARTICLE *particle;
        
        particle = (PARTICLE *)data;
        g_free(particle);
};

static void gps_write_particle(gpointer data, gpointer user_data) {
        PARTICLE *particle;
        FILE *output;
        
        particle = (PARTICLE *)data;
        output = (FILE *)user_data;
        
        fprintf(output, "%s\n", particle->name);
        fprintf(output, "%f\n", particle->mass);
        fprintf(output, "%f\n", particle->color[0]);
        fprintf(output, "%f\n", particle->color[1]);
        fprintf(output, "%f\n", particle->color[2]);
        fprintf(output, "%f\n", particle->position.x);
        fprintf(output, "%f\n", particle->position.y);
        fprintf(output, "%f\n", particle->position.z);
        fprintf(output, "%f\n", particle->velocity.x);
        fprintf(output, "%f\n", particle->velocity.y);
        fprintf(output, "%f\n", particle->velocity.z);
};



