/* GPS 0.1 - Gravitational Particle Simulator
 * Copyright (C) 2000 by Fabio Mancinelli <mancinel@univaq.it>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef HAVE_CONFIG_H
#include<config.h>
#endif

#include<GL/gl.h>
#include<GL/glu.h>
#include<gtkgl/gtkglarea.h>
#include<string.h>
#include<stdio.h>
#include<main_window.h>
#include<message_dialog.h>
#include<particle_dialog.h>
#include<particle_info_dialog.h>
#include<file_dialog.h>
#include<settings_dialog.h>
#include<vector_dialog.h>
#include<about_dialog.h>
#include<gps_gl.h>
#include<pixmaps/new.xpm>
#include<pixmaps/open.xpm>
#include<pixmaps/save.xpm>
#include<pixmaps/start.xpm>
#include<pixmaps/stop.xpm>
#include<pixmaps/step.xpm>

/******************************************************************************
  Static function prototypes
 ******************************************************************************/

static GtkWidget *menu_bar_new(GtkWidget *window, MAIN_WINDOW *main_window);
static GtkWidget *pixmap_widget_new(GtkWidget *window, gchar **xpm);
static GtkWidget *toolbar_new(GtkWidget *window, MAIN_WINDOW *main_window);
static GtkWidget *client_area_new(GtkWidget *window, MAIN_WINDOW *main_window);
static GtkWidget *statusbar_new(GtkWidget *window, MAIN_WINDOW *main_window);

static void draw_particle_iterator(gpointer data, gpointer user_data);
static void add_particle_iterator(gpointer data, gpointer user_data);

static void main_window_init(MAIN_WINDOW *main_window);
static void main_window_set_title(gchar *title, MAIN_WINDOW *main_window);
static void main_window_redraw_gl_area(MAIN_WINDOW *main_window);

static void pixmap_color_square_new(GtkWidget *window, GdkColor *color, GdkPixmap **pixmap, GdkPixmap **mask);
static void main_window_add_particle(PARTICLE *particle, MAIN_WINDOW *main_window);

/******************************************************************************
  Callback function prototypes
 ******************************************************************************/

static gint cb_delete(GtkWidget *widget, GdkEvent *event, gpointer data);
static gint cb_destroy(GtkWidget *widget, gpointer data);

static gint cb_gl_area_realize(GtkWidget *widget, gpointer data);
static gint cb_gl_area_expose(GtkWidget *widget, GdkEventExpose *event, gpointer data);
static gint cb_gl_area_configure(GtkWidget *widget, GdkEventConfigure *event, gpointer data);
static gint cb_gl_area_button_press(GtkWidget *widget, GdkEventButton *event, gpointer data);
static gint cb_gl_area_motion_notify(GtkWidget *widget, GdkEventMotion *event, gpointer data);

static void cb_new_item_activate(GtkWidget *widget, gpointer data);
static void cb_open_item_activate(GtkWidget *widget, gpointer data);
static void cb_save_item_activate(GtkWidget *widget, gpointer data);
static void cb_save_as_item_activate(GtkWidget *widget, gpointer data);
static void cb_exit_item_activate(GtkWidget *widget, gpointer data);

static void cb_show_grid_item_activate(GtkWidget *widget, gpointer data);
static void cb_show_axis_item_activate(GtkWidget *widget, gpointer data);
static void cb_set_camera_position_item_activate(GtkWidget *widget, gpointer data);
static void cb_set_target_position_item_activate(GtkWidget *widget, gpointer data);

static void cb_new_particle_item_activate(GtkWidget *widget, gpointer data);
static void cb_edit_particle_item_activate(GtkWidget *widget, gpointer data);
static void cb_delete_particle_item_activate(GtkWidget *widget, gpointer data);
static void cb_particle_info_item_activate(GtkWidget *widget, gpointer data);

static gint cb_timeout(gpointer data);
static void cb_start_item_activate(GtkWidget *widget, gpointer data);
static void cb_stop_item_activate(GtkWidget *widget, gpointer data);
static void cb_step_item_activate(GtkWidget *widget, gpointer data);
static void cb_settings_item_activate(GtkWidget *widget, gpointer data);

static void cb_about_item_activate(GtkWidget *widget, gpointer data);

static void cb_particles_clist_select_row(GtkWidget *widget, gint row, gint column, GdkEventButton *event, gpointer data);
static void cb_particles_clist_unselect_row(GtkWidget *widget, gint row, gint column, GdkEventButton *event, gpointer data);


/******************************************************************************
  Exported functions
 ******************************************************************************/

MAIN_WINDOW *main_window_new(gchar *title) {
        MAIN_WINDOW *main_window;
        GtkWidget *window;
        GtkWidget *vbox;
        GtkWidget *menu_bar;
        GtkWidget *toolbar;
        GtkWidget *client_area;
        GtkWidget *statusbar;
        
        /* Allocate memory for the main window structure */
        main_window = (MAIN_WINDOW *)g_malloc(sizeof(MAIN_WINDOW));
        
        /* Create the toplevel window */
        window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
        main_window->window = window;
        gtk_window_set_title(GTK_WINDOW(window), title);
        gtk_signal_connect(GTK_OBJECT(window), "delete_event", GTK_SIGNAL_FUNC(cb_delete), main_window);
        gtk_signal_connect(GTK_OBJECT(window), "destroy", GTK_SIGNAL_FUNC(cb_destroy), main_window);
        gtk_quit_add_destroy(TRUE, GTK_OBJECT(window));
        
        
        /* Create a vertical box to contain all the widgets */
        vbox = gtk_vbox_new(FALSE, 0);
        gtk_container_add(GTK_CONTAINER(window), vbox);
        
        /* Create the menu bar */
        menu_bar = menu_bar_new(window, main_window);
        gtk_widget_show(menu_bar);
        gtk_box_pack_start(GTK_BOX(vbox), menu_bar, FALSE, FALSE, 0);
        
        /* Create the toolbar */
        toolbar = toolbar_new(window, main_window);
        gtk_widget_show(toolbar);
        gtk_box_pack_start(GTK_BOX(vbox), toolbar, FALSE, FALSE, 0);
        
        /* Create the client area */
        client_area = client_area_new(window, main_window);
        gtk_widget_show(client_area);
        gtk_box_pack_start(GTK_BOX(vbox), client_area, TRUE, TRUE, 0);
        
        /* Create the statusbar */
        statusbar = statusbar_new(window, main_window);
        gtk_widget_show(statusbar);
        gtk_box_pack_start(GTK_BOX(vbox), statusbar, FALSE, FALSE, 0);
        
        gtk_widget_show(vbox);
        
        main_window_init(main_window);
                
        return main_window;
};

void main_window_show(MAIN_WINDOW *main_window) {
        gtk_widget_show(main_window->window);
};

void main_window_free(MAIN_WINDOW *main_window) {
        gps_free(main_window->gps);
        g_free(main_window);
};

/******************************************************************************
  Static functions
 ******************************************************************************/

static GtkWidget *menu_bar_new(GtkWidget *window, MAIN_WINDOW *main_window) {
        GtkWidget *menu_bar;
        GtkWidget *root_menu;        
        GtkWidget *menu;
        GtkWidget *menu_item;
        
        menu_bar = gtk_menu_bar_new();
        main_window->menu_bar = menu_bar;
        
        /* File menu */
        menu = gtk_menu_new();
        menu_item = gtk_menu_item_new_with_label("New");
        gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(cb_new_item_activate), main_window);
        gtk_widget_show(menu_item);
        gtk_menu_append(GTK_MENU(menu), menu_item);
        
        menu_item = gtk_menu_item_new_with_label("Open...");
        gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(cb_open_item_activate), main_window);
        gtk_widget_show(menu_item);
        gtk_menu_append(GTK_MENU(menu), menu_item);
        
        menu_item = gtk_menu_item_new_with_label("Save");
        gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(cb_save_item_activate), main_window);
        gtk_widget_show(menu_item);
        gtk_menu_append(GTK_MENU(menu), menu_item);
        
        menu_item = gtk_menu_item_new_with_label("Save as...");
        gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(cb_save_as_item_activate), main_window);
        gtk_widget_show(menu_item);
        gtk_menu_append(GTK_MENU(menu), menu_item);
        
        menu_item = gtk_menu_item_new();
        gtk_widget_show(menu_item);
        gtk_menu_append(GTK_MENU(menu), menu_item);
        
        menu_item = gtk_menu_item_new_with_label("Exit");
        gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(cb_exit_item_activate), main_window);
        gtk_widget_show(menu_item);
        gtk_menu_append(GTK_MENU(menu), menu_item);
        
        root_menu = gtk_menu_item_new_with_label("File");
        gtk_menu_item_set_submenu(GTK_MENU_ITEM(root_menu), menu);
        gtk_widget_show(root_menu);
        
        gtk_menu_bar_append(GTK_MENU_BAR(menu_bar), root_menu);

        /* View menu */
        menu = gtk_menu_new();
        menu_item = gtk_check_menu_item_new_with_label("Show grid");
        gtk_check_menu_item_set_show_toggle(GTK_CHECK_MENU_ITEM(menu_item), TRUE);
        gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menu_item), TRUE);
        gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(cb_show_grid_item_activate), main_window);        
        gtk_widget_show(menu_item);
        gtk_menu_append(GTK_MENU(menu), menu_item);
        
        menu_item = gtk_check_menu_item_new_with_label("Show axis");
        gtk_check_menu_item_set_show_toggle(GTK_CHECK_MENU_ITEM(menu_item), TRUE);
        gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menu_item), TRUE);
        gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(cb_show_axis_item_activate), main_window);
        gtk_widget_show(menu_item);
        gtk_menu_append(GTK_MENU(menu), menu_item);
        
        menu_item = gtk_menu_item_new();
        gtk_widget_show(menu_item);
        gtk_menu_append(GTK_MENU(menu), menu_item);
        
        menu_item = gtk_menu_item_new_with_label("Set camera position...");
        gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(cb_set_camera_position_item_activate), main_window);
        gtk_widget_show(menu_item);
        gtk_menu_append(GTK_MENU(menu), menu_item);
        
        menu_item = gtk_menu_item_new_with_label("Set target position...");
        gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(cb_set_target_position_item_activate), main_window);
        gtk_widget_show(menu_item);
        gtk_menu_append(GTK_MENU(menu), menu_item);
        
        root_menu = gtk_menu_item_new_with_label("View");
        gtk_menu_item_set_submenu(GTK_MENU_ITEM(root_menu), menu);
        gtk_widget_show(root_menu);
        
        gtk_menu_bar_append(GTK_MENU_BAR(menu_bar), root_menu);

        /* Particle menu */
        menu = gtk_menu_new();
        menu_item = gtk_menu_item_new_with_label("New particle...");
        gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(cb_new_particle_item_activate), main_window);
        gtk_widget_show(menu_item);
        gtk_menu_append(GTK_MENU(menu), menu_item);

        menu_item = gtk_menu_item_new_with_label("Edit particle...");
        gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(cb_edit_particle_item_activate), main_window);
        gtk_widget_show(menu_item);
        gtk_menu_append(GTK_MENU(menu), menu_item);
        
        menu_item = gtk_menu_item_new_with_label("Delete particle");
        gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(cb_delete_particle_item_activate), main_window);
        gtk_widget_show(menu_item);
        gtk_menu_append(GTK_MENU(menu), menu_item);
        
        menu_item = gtk_menu_item_new();
        gtk_widget_show(menu_item);
        gtk_menu_append(GTK_MENU(menu), menu_item);

        menu_item = gtk_menu_item_new_with_label("Particle info...");
        gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(cb_particle_info_item_activate), main_window);
        gtk_widget_show(menu_item);
        gtk_menu_append(GTK_MENU(menu), menu_item);



        root_menu = gtk_menu_item_new_with_label("Particle");
        gtk_menu_item_set_submenu(GTK_MENU_ITEM(root_menu), menu);
        gtk_widget_show(root_menu);
        
        gtk_menu_bar_append(GTK_MENU_BAR(menu_bar), root_menu);
        
        /* Simulation menu */
        menu = gtk_menu_new();
        menu_item = gtk_menu_item_new_with_label("Start");
        gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(cb_start_item_activate), main_window);
        gtk_widget_show(menu_item);
        gtk_menu_append(GTK_MENU(menu), menu_item);

        menu_item = gtk_menu_item_new_with_label("Stop");
        gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(cb_stop_item_activate), main_window);
        gtk_widget_show(menu_item);
        gtk_menu_append(GTK_MENU(menu), menu_item);
        
        menu_item = gtk_menu_item_new_with_label("Step");
        gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(cb_step_item_activate), main_window);
        gtk_widget_show(menu_item);
        gtk_menu_append(GTK_MENU(menu), menu_item);

        menu_item = gtk_menu_item_new();
        gtk_widget_show(menu_item);
        gtk_menu_append(GTK_MENU(menu), menu_item);
                
        menu_item = gtk_menu_item_new_with_label("Settings...");
        gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(cb_settings_item_activate), main_window);
        gtk_widget_show(menu_item);
        gtk_menu_append(GTK_MENU(menu), menu_item);
        
        root_menu = gtk_menu_item_new_with_label("Simulation");
        gtk_menu_item_set_submenu(GTK_MENU_ITEM(root_menu), menu);
        gtk_widget_show(root_menu);
        
        gtk_menu_bar_append(GTK_MENU_BAR(menu_bar), root_menu);
        
        /* Help menu */
        menu = gtk_menu_new();
        menu_item = gtk_menu_item_new_with_label("About...");
        gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(cb_about_item_activate), main_window);
        gtk_widget_show(menu_item);
        gtk_menu_append(GTK_MENU(menu), menu_item);

        root_menu = gtk_menu_item_new_with_label("Help");
        gtk_menu_item_set_submenu(GTK_MENU_ITEM(root_menu), menu);
        gtk_widget_show(root_menu);
        
        gtk_menu_bar_append(GTK_MENU_BAR(menu_bar), root_menu);

                                        
        return menu_bar;
};

static GtkWidget *pixmap_widget_new(GtkWidget *window, gchar **xpm) {
        GdkPixmap *pixmap;
        GdkPixmap *mask;
        GtkWidget *pixmap_widget;
        
        /* Create a pixmap widget for toolbar buttons */
        pixmap = gdk_pixmap_colormap_create_from_xpm_d(NULL, gtk_widget_get_colormap(window), &mask, NULL, xpm);
        pixmap_widget = gtk_pixmap_new(pixmap, mask);
        gdk_pixmap_unref(pixmap);
        gdk_pixmap_unref(mask);
        
        return pixmap_widget;
};

static GtkWidget *toolbar_new(GtkWidget *window, MAIN_WINDOW *main_window) {
        GtkWidget *toolbar;
        
        toolbar = gtk_toolbar_new(GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_BOTH);
        main_window->toolbar = toolbar;
        
        gtk_toolbar_append_item(GTK_TOOLBAR(toolbar), "New", "Create a new system", "New", pixmap_widget_new(window, new_xpm), cb_new_item_activate, main_window);
        gtk_toolbar_append_item(GTK_TOOLBAR(toolbar), "Open", "Open a saved system", "Open", pixmap_widget_new(window, open_xpm), cb_open_item_activate, main_window);
        gtk_toolbar_append_item(GTK_TOOLBAR(toolbar), "Save", "Save the current system", "Save", pixmap_widget_new(window, save_xpm), cb_save_item_activate, main_window);
        gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));
        gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));
        gtk_toolbar_append_item(GTK_TOOLBAR(toolbar), "Start", "Start the simulation", "Start", pixmap_widget_new(window, start_xpm), cb_start_item_activate, main_window);
        gtk_toolbar_append_item(GTK_TOOLBAR(toolbar), "Stop", "Stop the simulation", "Start", pixmap_widget_new(window, stop_xpm), cb_stop_item_activate, main_window);
        gtk_toolbar_append_item(GTK_TOOLBAR(toolbar), "Step", "Do the next simulation step", "Step", pixmap_widget_new(window, step_xpm), cb_step_item_activate, main_window);
               
        return toolbar;
        
};

static GtkWidget *client_area_new(GtkWidget *window, MAIN_WINDOW *main_window) {
        GtkWidget *hpaned;
        GtkWidget *particles_frame;
        GtkWidget *particles_frame_vbox;
        GtkWidget *gl_frame;
        GtkWidget *gl_frame_vbox;
        GtkWidget *gl_area;
        GtkWidget *particles_clist;
        
        gchar *titles[2] = {"", "Name"};
        
        int attr_list[] = {
                GDK_GL_RGBA,
                GDK_GL_RED_SIZE, 1,
                GDK_GL_GREEN_SIZE, 1,
                GDK_GL_BLUE_SIZE, 1,
                GDK_GL_DOUBLEBUFFER, 
                GDK_GL_NONE
        };
        
        /* Create an horizontal paned area */
        hpaned = gtk_hpaned_new();
        
        /* Create the particle list and its frame */
        particles_frame = gtk_frame_new("Particles");
        gtk_container_set_border_width(GTK_CONTAINER(particles_frame), 10);
        
        particles_frame_vbox = gtk_vbox_new(FALSE, 0);
        gtk_container_set_border_width(GTK_CONTAINER(particles_frame_vbox), 5);
                
        particles_clist = gtk_clist_new_with_titles(2, titles);
        main_window->particles_clist = particles_clist;
        gtk_clist_set_column_width(GTK_CLIST(particles_clist), 0, 12);
        gtk_clist_set_column_resizeable(GTK_CLIST(particles_clist), 0, FALSE);
        gtk_clist_set_column_width(GTK_CLIST(particles_clist), 1, 150);
        gtk_signal_connect(GTK_OBJECT(particles_clist), "select_row", GTK_SIGNAL_FUNC(cb_particles_clist_select_row), main_window);
        gtk_signal_connect(GTK_OBJECT(particles_clist), "unselect_row", GTK_SIGNAL_FUNC(cb_particles_clist_unselect_row), main_window);
        gtk_widget_show(particles_clist);
        gtk_box_pack_start(GTK_BOX(particles_frame_vbox), particles_clist, TRUE, TRUE, 0);
        
        gtk_widget_show(particles_frame_vbox);
        gtk_container_add(GTK_CONTAINER(particles_frame), particles_frame_vbox);
        
        gtk_widget_show(particles_frame);  
        gtk_paned_add1(GTK_PANED(hpaned), particles_frame); 
        
        /* Create the 3D area and its frame */
        gl_frame = gtk_frame_new("3D");
        gtk_container_set_border_width(GTK_CONTAINER(gl_frame), 10);

        gl_frame_vbox = gtk_vbox_new(FALSE, 0);
        gtk_container_set_border_width(GTK_CONTAINER(gl_frame_vbox), 5);
               
        gl_area = gtk_gl_area_new(attr_list);
        main_window->gl_area = gl_area;
                
        gtk_widget_set_usize(gl_area, 800, 600);
        gtk_widget_set_events(gl_area, GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK);
        gtk_signal_connect(GTK_OBJECT(gl_area), "realize", GTK_SIGNAL_FUNC(cb_gl_area_realize), main_window);
        gtk_signal_connect(GTK_OBJECT(gl_area), "expose_event", GTK_SIGNAL_FUNC(cb_gl_area_expose), main_window);
        gtk_signal_connect(GTK_OBJECT(gl_area), "configure_event", GTK_SIGNAL_FUNC(cb_gl_area_configure), main_window);
        gtk_signal_connect(GTK_OBJECT(gl_area), "button_press_event", GTK_SIGNAL_FUNC(cb_gl_area_button_press), main_window);
        gtk_signal_connect(GTK_OBJECT(gl_area), "motion_notify_event", GTK_SIGNAL_FUNC(cb_gl_area_motion_notify), main_window);
        gtk_widget_show(gl_area);
        gtk_box_pack_start(GTK_BOX(gl_frame_vbox), gl_area, TRUE, TRUE, 0);                        

        gtk_widget_show(gl_frame_vbox);
        gtk_container_add(GTK_CONTAINER(gl_frame), gl_frame_vbox);
        
        gtk_widget_show(gl_frame);        
        gtk_paned_add2(GTK_PANED(hpaned), gl_frame);
        
        return hpaned;
};

static GtkWidget *statusbar_new(GtkWidget *window, MAIN_WINDOW *main_window) {
        GtkWidget *statusbar;
        
        statusbar = gtk_statusbar_new();        
        main_window->statusbar = statusbar;
        
        return statusbar;
};        

static void draw_particle_iterator(gpointer data, gpointer user_data) {
        PARTICLE *particle;
        
        particle = (PARTICLE *)data;
        gps_gl_draw_particle(particle);
};

static void add_particle_iterator(gpointer data, gpointer user_data) {
        MAIN_WINDOW *main_window;
        PARTICLE *particle;
        
        main_window = (MAIN_WINDOW *)user_data;
        particle = (PARTICLE *)data;
        
        main_window_add_particle(particle, main_window);
};


static void main_window_init(MAIN_WINDOW *main_window) {
                main_window->show_grid = TRUE;
                main_window->show_axis = TRUE;

                main_window->gps = gps_new();
                main_window->gps_saved = FALSE;
                main_window->gps_file_name[0] = '\0';
                main_window->gps_modified = FALSE;
                
                main_window->selected_particle = NULL;
                
                main_window->running = FALSE;
                main_window->timeout_tag = 0;
                main_window->update_interval = 50;
                
                main_window_set_title(NULL, main_window);
                main_window_redraw_gl_area(main_window);
                
                gtk_clist_clear(GTK_CLIST(main_window->particles_clist));        
};

static void main_window_set_title(gchar *title, MAIN_WINDOW *main_window) {
        static gchar buffer[MAX_FILE_NAME_LENGTH];
        
        if(title == NULL) strcpy(buffer, "GPS");
        else {
                strcpy(buffer, "GPS - ");
                strcat(buffer, title);
        };
        
        gtk_window_set_title(GTK_WINDOW(main_window->window), buffer);
};

static void main_window_redraw_gl_area(MAIN_WINDOW *main_window) {
        GdkRectangle area;
               
        area.x = 0;
        area.y = 0;
        area.width = main_window->gl_area->allocation.width;
        area.height = main_window->gl_area->allocation.height;
        
        gtk_widget_draw(main_window->gl_area, &area);
};

static void pixmap_color_square_new(GtkWidget *window, GdkColor *color, GdkPixmap **pixmap, GdkPixmap **mask) {
        static char *color_square_xpm[] = {
                "10 10 3 1",
                "       c None",
                "#      c #000000",
                ".      c #FFFFFF",
                "##########",
                "#........#",
                "#........#",
                "#........#",
                "#........#",
                "#........#",
                "#........#",
                "#........#",
                "#........#",
                "##########"
        };
        static char buffer[32];
        
        sprintf(buffer, ".\tc #%2X%2X%2X", color->red, color->green, color->blue);
        color_square_xpm[3] = buffer;
        *pixmap = gdk_pixmap_colormap_create_from_xpm_d(NULL, gtk_widget_get_colormap(window), mask, NULL, color_square_xpm);
};

static void main_window_add_particle(PARTICLE *particle, MAIN_WINDOW *main_window) {
        GdkPixmap *pixmap;
        GdkPixmap *mask;
        GdkColor color;
        gchar *item[2];
        int row;
        
        item[0] = particle->name;
        item[1] = particle->name;
        
        row = gtk_clist_append(GTK_CLIST(main_window->particles_clist), item); 
        color.red = (gushort)(particle->color[0] * 255.0);
        color.green = (gushort)(particle->color[1] * 255.0);
        color.blue = (gushort)(particle->color[2] * 255.0);
        pixmap_color_square_new(main_window->window, &color, &pixmap, &mask);
        gtk_clist_set_pixmap(GTK_CLIST(main_window->particles_clist), row, 0, pixmap, mask);
        gtk_clist_set_row_data(GTK_CLIST(main_window->particles_clist), row, particle);
};
        
                

/******************************************************************************
  Callback functions
 ******************************************************************************/

static gint cb_delete(GtkWidget *widget, GdkEvent *event, gpointer data) {
        MAIN_WINDOW *main_window;
        MESSAGE_DIALOG *message_dialog;
        MESSAGE_DIALOG_DATA *message_dialog_data;
        gint to_ret;
        
        main_window = (MAIN_WINDOW *)data;
        
        /* By default, don't ask anything */
        to_ret = FALSE;
        
        /* If there is something to save show the "are you sure" dialog */
        if(main_window->gps_modified == TRUE) {                
                message_dialog = message_dialog_new("Warning", "There are unsaved data. Are you sure you want to quit?", YES_BUTTON | NO_BUTTON);
                message_dialog_show(message_dialog);
                message_dialog_data = message_dialog_get_data(message_dialog);

                if(message_dialog_data->result == ID_YES) to_ret = FALSE;
                else to_ret = TRUE;

                message_dialog_free(message_dialog);
        }
                
        return to_ret;        
};
        
static gint cb_destroy(GtkWidget *widget, gpointer data) {
        MAIN_WINDOW *main_window;
        
        main_window = (MAIN_WINDOW *)data;
        
        if(main_window->running == TRUE) gtk_timeout_remove(main_window->timeout_tag);
                        
        /* Quit the main loop */
        gtk_main_quit();
        
        return TRUE;
};

static gint cb_gl_area_realize(GtkWidget *widget, gpointer data) {
        MAIN_WINDOW *main_window;
        static GLfloat pos[4] = {5.0, 5.0, 10.0, 0.0};
        static GLfloat white[4] = {0.8, 0.8, 0.8, 1.0};        
                
        main_window = (MAIN_WINDOW *)data;
        
        /* Initialize OpenGL */
        if (gtk_gl_area_begingl(GTK_GL_AREA(widget))) {
                glEnable(GL_CULL_FACE);
                glEnable(GL_DEPTH_TEST);
                
                glShadeModel(GL_SMOOTH);  
                
                glEnable(GL_LIGHTING);
                glLightfv(GL_LIGHT0, GL_POSITION, pos);
                glEnable(GL_LIGHT0);
                glEnable(GL_LIGHT_MODEL_TWO_SIDE);
                glEnable(GL_NORMALIZE);
                
                glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, white); 
                
                /* Create the 3D grid */
                main_window->grid = gps_gl_xz_grid(-50, 50, -50, 50, -1.0, 50, 50);
                
                /* Create the axis */
                main_window->axis = gps_gl_axis(0.05, 1.0, 0.1, 0.2);                                
                gtk_gl_area_endgl(GTK_GL_AREA(widget));
        };               

        return TRUE;
};


static gint cb_gl_area_expose(GtkWidget *widget, GdkEventExpose *event, gpointer data) {
        MAIN_WINDOW *main_window;

        main_window = (MAIN_WINDOW *)data;
        
        /* Draw only last expose */
        if (event->count > 0) return TRUE;
        
        if (gtk_gl_area_begingl(GTK_GL_AREA(widget))) {
                /* Clear the area */
                glClearColor(0.0, 0.0, 0.0, 1.0);
                glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
                
                glMatrixMode(GL_PROJECTION);
                glLoadIdentity();
                gluPerspective(main_window->gps->fov, 1.3, 1.0, 100.0);
                
                glMatrixMode(GL_MODELVIEW);
                glLoadIdentity();
                gluLookAt(main_window->gps->camera.x, main_window->gps->camera.y, main_window->gps->camera.z, main_window->gps->target.x, main_window->gps->target.y, main_window->gps->target.z, 0.0, 1.0, 0.0);
                
                /* Display 3D grid */
                if(main_window->show_grid == TRUE) glCallList(main_window->grid);
                
                /* Display axis */
                if(main_window->show_axis == TRUE) glCallList(main_window->axis);
                                
                /* Draw particles */
                g_list_foreach(main_window->gps->particles_list, draw_particle_iterator, NULL);

                /* Draw the bounding box of the selected particle */
                if(main_window->selected_particle != NULL) {
                        gps_gl_draw_particle_bounding_box(main_window->selected_particle);
                };

                /* Swap backbuffer to front */
                gtk_gl_area_swapbuffers(GTK_GL_AREA(widget));
                
                gtk_gl_area_begingl(GTK_GL_AREA(widget));
        };

        return TRUE;
};
                
static gint cb_gl_area_configure(GtkWidget *widget, GdkEventConfigure *event, gpointer data) {        
        /* Update the viewport and set it to the new widget size */
        if (gtk_gl_area_begingl(GTK_GL_AREA(widget))) {
                glViewport(0, 0, (GLint)widget->allocation.width, (GLint)widget->allocation.height);
                gtk_gl_area_endgl(GTK_GL_AREA(widget));
        };
        
        return TRUE;
};

static gint cb_gl_area_button_press(GtkWidget *widget, GdkEventButton *event, gpointer data) {
        MAIN_WINDOW *main_window;
        
        main_window = (MAIN_WINDOW *)data;
        
        /* If a button is pressed in the 3D area then a dragging operation has been started */
        if(event->button == 1) {
                main_window->dragging.x = event->x;
                main_window->dragging.y = event->y;
                
                return TRUE;
        };
        
        return FALSE;
};

static gint cb_gl_area_motion_notify(GtkWidget *widget, GdkEventMotion *event, gpointer data) {
        MAIN_WINDOW *main_window;
        int x;
        int y;        
        GdkModifierType state;
        GdkRectangle area;
        VECTOR delta;
        
        main_window = (MAIN_WINDOW *)data;
        
        if(event->is_hint) {
                gdk_window_get_pointer(event->window, &x, &y, &state);
        }
        else {
                x = event->x;
                y = event->y;
                state = event->state;
        };
        
        /* Compute the delta : current position - old dragging position */
        delta.x = (x - main_window->dragging.x) / 3.0;
        delta.y = (y - main_window->dragging.y) / 3.0;
        
        /* Set the current dragging position as the old dragging position */
        main_window->dragging.x = x;
        main_window->dragging.y = y;
        
        /* This is used to repaint the entire 3D area */
        area.x = 0;
        area.y = 0;
        area.width = widget->allocation.width;
        area.height = widget->allocation.height;
        
        if(state & GDK_SHIFT_MASK) {                
                if(state & GDK_BUTTON1_MASK) {
                        /* The user is dragging using the first button and the SHIFT key. Update target xy position */
                        main_window->gps->target.x -= delta.x;
                        main_window->gps->target.y += delta.y;
                
                        gtk_widget_draw(widget, &area);                
                }
                else if(state & GDK_BUTTON3_MASK) {
                        /* The user is dragging using the third button and the SHIFT key. Update target zy position */
                        main_window->gps->target.z -= delta.x;
                        main_window->gps->target.y += delta.y;
                
                        gtk_widget_draw(widget, &area);                
                };
        }
        else {
                if(state & GDK_BUTTON1_MASK) {
                        /* The user is dragging using the first button. Update camera xy position */
                        main_window->gps->camera.x -= delta.x;
                        main_window->gps->camera.y += delta.y;

                        gtk_widget_draw(widget, &area);
                }
                else if(state & GDK_BUTTON3_MASK) {
                        /* The user is dragging using the third button. Update camera zy position */
                        main_window->gps->camera.z -= delta.x;
                        main_window->gps->camera.y += delta.y;

                        gtk_widget_draw(widget, &area);
                };
        };
                
        return TRUE;
};

static void cb_new_item_activate(GtkWidget *widget, gpointer data) {
        MAIN_WINDOW *main_window;
        MESSAGE_DIALOG *message_dialog;
        MESSAGE_DIALOG_DATA *message_dialog_data;
        gint ok;
        
        main_window = (MAIN_WINDOW *)data;
        
        ok = TRUE;
        
        /* Check if there are unsaved changes */
        if(main_window->gps_modified == TRUE) {
                message_dialog = message_dialog_new("Warning", "There are unsaved data. Are you sure you want to quit?", YES_BUTTON | NO_BUTTON);
                message_dialog_show(message_dialog);
                message_dialog_data = message_dialog_get_data(message_dialog);

                if(message_dialog_data->result == ID_YES) ok = TRUE;
                else ok = FALSE;

                message_dialog_free(message_dialog);
        };
        
        /* If the user wishes to continue then initialize all the environment */
        if(ok == TRUE) {                                
                gps_free(main_window->gps);
                main_window_init(main_window);
        };
};

static void cb_open_item_activate(GtkWidget *widget, gpointer data) {
        MAIN_WINDOW *main_window;
        FILE_DIALOG *file_dialog;
        FILE_DIALOG_DATA *file_dialog_data;
        MESSAGE_DIALOG *message_dialog;
        MESSAGE_DIALOG_DATA *message_dialog_data;
        GPS *gps;
        gint ok;
        
        main_window = (MAIN_WINDOW *)data;
        
        ok = TRUE;
        
        /* Check if there are unsaved changes */
        if(main_window->gps_modified == TRUE) {
                message_dialog = message_dialog_new("Warning", "There are unsaved data. Are you sure you want to continue?", YES_BUTTON | NO_BUTTON);
                message_dialog_show(message_dialog);
                message_dialog_data = message_dialog_get_data(message_dialog);

                if(message_dialog_data->result == ID_YES) ok = TRUE;
                else ok = FALSE;

                message_dialog_free(message_dialog);
        };
        
        /* If the user doesn't wanto to continue then return */
        if(ok == FALSE) return;        
        
        /* Open a file dialog to get the file name */                        
        file_dialog = file_dialog_new("Open...", "*.gps");
        file_dialog_show(file_dialog);
        file_dialog_data = file_dialog_get_data(file_dialog);
        
        
        if(file_dialog_data->result == ID_OK) {
                /* Create a new GPS structure */
                gps = gps_new();
                
                /* Load the saved system into the newly created structure */
                if(gps_load(file_dialog_data->file_name, gps) == FALSE) {
                        /* If something goes wrong then return */
                        message_dialog = message_dialog_new("Error", "Unable to load file", OK_BUTTON);
                        message_dialog_show(message_dialog);
                        message_dialog_free(message_dialog);
                        
                        gps_free(gps);
                }
                else {    
                        /* Free the current GPS structure */
                        gps_free(main_window->gps);
                        
                        /* Initialize the environment */
                        main_window_init(main_window);
                        g_list_foreach(gps->particles_list, add_particle_iterator, main_window);

                        /* Set the current GPS structure to the loaded system */
                        main_window->gps = gps;
                        
                        /* General setup */
                        main_window->gps_saved = TRUE;
                        strcpy(main_window->gps_file_name, file_dialog_data->file_name);
                        main_window->gps_modified = FALSE;
                        
                        
                        
                        main_window_set_title(main_window->gps_file_name, main_window);
                        main_window_redraw_gl_area(main_window);
                };
        

        };
                
        file_dialog_free(file_dialog);
};

static void cb_save_item_activate(GtkWidget *widget, gpointer data) {
        MAIN_WINDOW *main_window;
        MESSAGE_DIALOG *message_dialog;
        
        main_window = (MAIN_WINDOW *)data;
        
        /* If we have a previously loaded system then overwrite that system  else act as a save as */
        if(main_window->gps_saved == TRUE) {
                if(gps_save(main_window->gps_file_name, main_window->gps) == FALSE) {
                        message_dialog = message_dialog_new("Error", "Unable to save file", OK_BUTTON);
                        message_dialog_show(message_dialog);
                        message_dialog_free(message_dialog);
                }
                else {
                        main_window->gps_modified = FALSE;                        
                };
        }
        else cb_save_as_item_activate(widget, data);
};

static void cb_save_as_item_activate(GtkWidget *widget, gpointer data) {
        MAIN_WINDOW *main_window;
        FILE_DIALOG *file_dialog;
        FILE_DIALOG_DATA *file_dialog_data;
        MESSAGE_DIALOG *message_dialog;
        
        main_window = (MAIN_WINDOW *)data;
        
        /* Open a file dialog to get the file name */
        file_dialog = file_dialog_new("Save as...", "*.gps");
        file_dialog_show(file_dialog);
        file_dialog_data = file_dialog_get_data(file_dialog);
        
        if(file_dialog_data->result == ID_OK) {
                if(gps_save(file_dialog_data->file_name, main_window->gps) == FALSE) {
                        message_dialog = message_dialog_new("Error", "Unable to save file", OK_BUTTON);
                        message_dialog_show(message_dialog);
                        message_dialog_free(message_dialog);
                }
                else {
                        /* Update variables */
                        main_window->gps_saved = TRUE;
                        strcpy(main_window->gps_file_name, file_dialog_data->file_name);
                        main_window->gps_modified = FALSE;
                        
                        main_window_set_title(main_window->gps_file_name, main_window);
                };
        };
        
        file_dialog_free(file_dialog);
};

static void cb_exit_item_activate(GtkWidget *widget, gpointer data) {
        MAIN_WINDOW *main_window;
        MESSAGE_DIALOG *message_dialog;
        MESSAGE_DIALOG_DATA *message_dialog_data;
        gint ok;
        
        main_window = (MAIN_WINDOW *)data;
                
        ok = TRUE;                
                
        /* If there is something to save show the "are you sure" dialog */
        if(main_window->gps_modified == TRUE) {                
                message_dialog = message_dialog_new("Warning", "There are unsaved data. Are you sure you want to quit?", YES_BUTTON | NO_BUTTON);
                message_dialog_show(message_dialog);
                message_dialog_data = message_dialog_get_data(message_dialog);

                if(message_dialog_data->result == ID_YES) ok = TRUE;
                else ok = FALSE;

                message_dialog_free(message_dialog);
        };
        
        if(ok == TRUE) gtk_widget_destroy(main_window->window);
};        

static void cb_show_grid_item_activate(GtkWidget *widget, gpointer data) {
        MAIN_WINDOW *main_window;
        
        main_window = (MAIN_WINDOW *)data;
        
        if(main_window->show_grid == TRUE) main_window->show_grid = FALSE;
        else main_window->show_grid = TRUE;        
        
        main_window_redraw_gl_area(main_window);
};

static void cb_show_axis_item_activate(GtkWidget *widget, gpointer data) {
        MAIN_WINDOW *main_window;
        
        main_window = (MAIN_WINDOW *)data;
        
        if(main_window->show_axis == TRUE) main_window->show_axis = FALSE;
        else main_window->show_axis = TRUE;        
        
        main_window_redraw_gl_area(main_window);
};

static void cb_set_camera_position_item_activate(GtkWidget *widget, gpointer data) {
        MAIN_WINDOW *main_window;
        VECTOR_DIALOG *vector_dialog;
        VECTOR_DIALOG_DATA *vector_dialog_data;
        
        main_window = (MAIN_WINDOW *)data;
        
        vector_dialog = vector_dialog_new("Camera position", &main_window->gps->camera);
        vector_dialog_show(vector_dialog);
        
        vector_dialog_data = vector_dialog_get_data(vector_dialog);
        if(vector_dialog_data->result == ID_OK) {
                vector_copy(&main_window->gps->camera, &vector_dialog_data->v);
                
                main_window_redraw_gl_area(main_window);
        };
        
        vector_dialog_free(vector_dialog);        
};

static void cb_set_target_position_item_activate(GtkWidget *widget, gpointer data) {
        MAIN_WINDOW *main_window;
        VECTOR_DIALOG *vector_dialog;
        VECTOR_DIALOG_DATA *vector_dialog_data;
        
        main_window = (MAIN_WINDOW *)data;
        
        vector_dialog = vector_dialog_new("Target position", &main_window->gps->target);
        vector_dialog_show(vector_dialog);
        
        vector_dialog_data = vector_dialog_get_data(vector_dialog);
        if(vector_dialog_data->result == ID_OK) {
                vector_copy(&main_window->gps->target, &vector_dialog_data->v);
                
                main_window_redraw_gl_area(main_window);
        };
        
        vector_dialog_free(vector_dialog);
};


static void cb_new_particle_item_activate(GtkWidget *widget, gpointer data) {
        MAIN_WINDOW *main_window;
        PARTICLE_DIALOG *particle_dialog;
        PARTICLE_DIALOG_DATA *particle_dialog_data;
        PARTICLE *particle;
        
        main_window = (MAIN_WINDOW *)data;
        
        particle_dialog = particle_dialog_new("New particle", NULL);
        particle_dialog_show(particle_dialog);
        
        particle_dialog_data = particle_dialog_get_data(particle_dialog);
        if(particle_dialog_data->result == ID_OK) {
                /* Create a new particle */
                particle = (PARTICLE *)g_malloc(sizeof(PARTICLE));
                memcpy(particle, &particle_dialog_data->particle, sizeof(PARTICLE));
                
                gps_add_particle(particle, main_window->gps);                                
                main_window_add_particle(particle, main_window);
                
                main_window->gps_modified = TRUE;
                
                main_window_redraw_gl_area(main_window);
        };
                

        
        particle_dialog_free(particle_dialog);
};

static void cb_edit_particle_item_activate(GtkWidget *widget, gpointer data) {
        MAIN_WINDOW *main_window;
        PARTICLE_DIALOG *particle_dialog;
        PARTICLE_DIALOG_DATA *particle_dialog_data;
        MESSAGE_DIALOG *message_dialog;
        GdkPixmap *pixmap;
        GdkPixmap *mask;
        GdkColor color;
        int row;        
        
        main_window = (MAIN_WINDOW *)data;
        
        if(main_window->selected_particle == NULL) {
                message_dialog = message_dialog_new("Error", "No selected particle", OK_BUTTON);
                message_dialog_show(message_dialog);
                message_dialog_free(message_dialog);
                
                return;
        };        
                        
        particle_dialog = particle_dialog_new("Edit particle", main_window->selected_particle);
        particle_dialog_show(particle_dialog);
        
        particle_dialog_data = particle_dialog_get_data(particle_dialog);
        if(particle_dialog_data->result == ID_OK) {
                /* Update the changed values */
                strcpy(main_window->selected_particle->name, particle_dialog_data->particle.name);
                main_window->selected_particle->mass = particle_dialog_data->particle.mass;
                main_window->selected_particle->color[0] = particle_dialog_data->particle.color[0];
                main_window->selected_particle->color[1] = particle_dialog_data->particle.color[1];
                main_window->selected_particle->color[2] = particle_dialog_data->particle.color[2];
                                
                /* Change the color and the name in the clist */
                row = gtk_clist_find_row_from_data(GTK_CLIST(main_window->particles_clist), main_window->selected_particle); 
                color.red = (gushort)(main_window->selected_particle->color[0] * 255.0);
                color.green = (gushort)(main_window->selected_particle->color[1] * 255.0);
                color.blue = (gushort)(main_window->selected_particle->color[2] * 255.0);
                pixmap_color_square_new(main_window->window, &color, &pixmap, &mask);
                gtk_clist_set_pixmap(GTK_CLIST(main_window->particles_clist), row, 0, pixmap, mask);
                gtk_clist_set_text(GTK_CLIST(main_window->particles_clist), row, 1, particle_dialog_data->particle.name);

                vector_copy(&main_window->selected_particle->position, &particle_dialog_data->particle.position);
                vector_copy(&main_window->selected_particle->velocity, &particle_dialog_data->particle.velocity);
                
                main_window->gps_modified = TRUE;
                
                main_window_redraw_gl_area(main_window);
        };
                
        particle_dialog_free(particle_dialog);
};

static void cb_delete_particle_item_activate(GtkWidget *widget, gpointer data) {
        MAIN_WINDOW *main_window;
        MESSAGE_DIALOG *message_dialog;
        MESSAGE_DIALOG_DATA *message_dialog_data;
        gint row;
        
        main_window = (MAIN_WINDOW *)data;
        
        if(main_window->selected_particle == NULL) {
                message_dialog = message_dialog_new("Error", "No selected particle", OK_BUTTON);
                message_dialog_show(message_dialog);
                message_dialog_free(message_dialog);
                
                return;
        };      
        
        message_dialog = message_dialog_new("Warning", "Are you sure?", YES_BUTTON | NO_BUTTON);
        message_dialog_show(message_dialog);
        
        message_dialog_data = message_dialog_get_data(message_dialog);
        
        if(message_dialog_data->result == ID_YES) {
                
                row = gtk_clist_find_row_from_data(GTK_CLIST(main_window->particles_clist), main_window->selected_particle);
                gps_delete_particle(main_window->selected_particle, main_window->gps);
                gtk_clist_remove(GTK_CLIST(main_window->particles_clist), row);
                
                main_window->gps_modified = TRUE;
                
                main_window_redraw_gl_area(main_window);
        };
        
        message_dialog_free(message_dialog);
};

static void cb_particle_info_item_activate(GtkWidget *widget, gpointer data) {
        MAIN_WINDOW *main_window;
        MESSAGE_DIALOG *message_dialog;
        PARTICLE_INFO_DIALOG *particle_info_dialog;
        
        main_window = (MAIN_WINDOW *)data;
        
        if(main_window->selected_particle == NULL) {
                message_dialog = message_dialog_new("Error", "No selected particle", OK_BUTTON);
                message_dialog_show(message_dialog);
                message_dialog_free(message_dialog);
                
                return;
        };      
        
        particle_info_dialog = particle_info_dialog_new("Particle info", main_window->selected_particle);
        particle_info_dialog_show(particle_info_dialog);
        particle_info_dialog_free(particle_info_dialog);
};

static gint cb_timeout(gpointer data) {
        MAIN_WINDOW *main_window;
        
        main_window = (MAIN_WINDOW *)data;
        gps_update(main_window->gps);

        main_window_redraw_gl_area(main_window);
        
        return TRUE;
};

static void cb_start_item_activate(GtkWidget *widget, gpointer data) {
        MAIN_WINDOW *main_window;
        
        main_window = (MAIN_WINDOW *)data;
        
        if(main_window->running == TRUE) return;
        
        main_window->timeout_tag = gtk_timeout_add(main_window->update_interval, cb_timeout, main_window);
        main_window->running = TRUE;        
};

static void cb_stop_item_activate(GtkWidget *widget, gpointer data) {
        MAIN_WINDOW *main_window;
        
        main_window = (MAIN_WINDOW *)data;
        
        if(main_window->running == FALSE) return;
        
        gtk_timeout_remove(main_window->timeout_tag);
        main_window->running = FALSE;

};

static void cb_step_item_activate(GtkWidget *widget, gpointer data) {
        MAIN_WINDOW *main_window;
        
        main_window = (MAIN_WINDOW *)data;
        
        if(main_window->running == TRUE) return;
        cb_timeout(main_window);

};

static void cb_settings_item_activate(GtkWidget *widget, gpointer data) {
        MAIN_WINDOW *main_window;
        SETTINGS_DIALOG *settings_dialog;
        SETTINGS_DIALOG_DATA *settings_dialog_data;
        
        main_window = (MAIN_WINDOW *)data;
        
        settings_dialog = settings_dialog_new(main_window->gps);
        settings_dialog_show(settings_dialog);
        
        settings_dialog_data = settings_dialog_get_data(settings_dialog);
        
        if(settings_dialog_data->result == ID_OK) {
                main_window->gps->g = settings_dialog_data->g;
                main_window->gps->step_size = settings_dialog_data->step_size;
                main_window->gps->variable_step_size = settings_dialog_data->variable_step_size;
                main_window->gps->max_step_size = settings_dialog_data->max_step_size;
                main_window->gps->separation = settings_dialog_data->separation;
        };
        
        settings_dialog_free(settings_dialog);
};

static void cb_about_item_activate(GtkWidget *widget, gpointer data) {
        MAIN_WINDOW *main_window;
        ABOUT_DIALOG *about_dialog;

        
        main_window = (MAIN_WINDOW *)data;
        
        about_dialog = about_dialog_new();
        about_dialog_show(about_dialog);
        about_dialog_free(about_dialog);
};

static void cb_particles_clist_select_row(GtkWidget *widget, gint row, gint column, GdkEventButton *event, gpointer data) {
        MAIN_WINDOW *main_window;
        PARTICLE *particle;
        
        main_window = (MAIN_WINDOW *)data;
        particle = gtk_clist_get_row_data(GTK_CLIST(main_window->particles_clist), row);
        main_window->selected_particle = particle;
        
        main_window_redraw_gl_area(main_window);
};

static void cb_particles_clist_unselect_row(GtkWidget *widget, gint row, gint column, GdkEventButton *event, gpointer data) {
        MAIN_WINDOW *main_window;
                
        main_window = (MAIN_WINDOW *)data;
        main_window->selected_particle = NULL;
        
        main_window_redraw_gl_area(main_window);
};

