/* GPS 0.1 - Gravitational Particle Simulator
 * Copyright (C) 2000 by Fabio Mancinelli <mancinel@univaq.it>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef HAVE_CONFIG_H
#include<config.h>
#endif

#include<stdlib.h>
#include<math.h>
#include<rk.h>



/******************************************************************************
  Static function prototypes
 ******************************************************************************/

static void update_min_distance(PARTICLE *p0, GList *particles_list);
static void update_acceleration(PARTICLE *p0, GList *particles_list, double g);
static void update_positions(GList *particles_list, double h, double g);
static int check_step(GList *particles_list, double k);


        
/******************************************************************************
  Exported functions
 ******************************************************************************/

void rk_step(GPS *gps) {
        PARTICLE *particle;
        int i;
        
        if(gps->variable_step_size == FALSE) {
                update_positions(gps->particles_list, gps->step_size, gps->g);
        }
        else {
                /* For every particle save initial position and velocity */        
                for(i = 0; i < g_list_length(gps->particles_list); i++) {
                        particle = (PARTICLE *)g_list_nth_data(gps->particles_list, i);
                        vector_copy(&particle->v0, &particle->velocity);
                        vector_copy(&particle->p0, &particle->position);
                };

                /* For every particle compute its minimum distance from another particle */
                for(i = 0; i < g_list_length(gps->particles_list); i++) {
                        particle = (PARTICLE *)g_list_nth_data(gps->particles_list, i);
                        update_min_distance(particle, gps->particles_list);
                };

                while(TRUE) {
                        update_positions(gps->particles_list, gps->step_size, gps->g);

                        if(check_step(gps->particles_list, gps->separation)) {
                                /* Everything moved well, so we can increment step size */
                                gps->step_size *= 1.5;
                                if(gps->step_size > gps->max_step_size) gps->step_size = gps->max_step_size;
                                break;
                        }
                        else {
                                /* Something moved too fast, adjust step size */
                                gps->step_size /= 2;

                                /* For every particle restore initial position and velocity */
                                for(i = 0; i < g_list_length(gps->particles_list); i++) {
                                        particle = (PARTICLE *)g_list_nth_data(gps->particles_list, i);
                                        vector_copy(&particle->velocity, &particle->v0);
                                        vector_copy(&particle->position, &particle->p0);
                                };
                        };
                };
        };        
};
                        


/******************************************************************************
  Static functions
 ******************************************************************************/

static void update_min_distance(PARTICLE *p0, GList *particles_list) {        
        PARTICLE *p1;
        int i;
        double min_distance;
        
        min_distance = (double)ULONG_MAX;
        
        for(i = 0; i < g_list_length(particles_list); i++) {
                p1 = (PARTICLE *)g_list_nth_data(particles_list, i);
                if(p0 != p1) {
                        VECTOR tmp;
                        double d;
                        
                        vector_sub(&tmp, &p1->position, &p0->position);
                        d = vector_module(&tmp);
                        if(d < min_distance) min_distance = d;
                };
        };
        
        p0->min_distance = min_distance;
};                        

        
static void update_acceleration(PARTICLE *p0, GList *particles_list, double g) {
        PARTICLE *p1;
        VECTOR f;
        int i;
        
        vector_set(&f, 0.0, 0.0, 0.0);
        
        for(i = 0; i < g_list_length(particles_list); i++) {
                p1 = (PARTICLE *)g_list_nth_data(particles_list, i);
                if(p0 != p1) {
                        VECTOR tmp;
                        double r3;
                        
                        vector_sub(&tmp, &p1->position, &p0->position);
                        r3 = pow(vector_module(&tmp), 3);
                        vector_scale(&tmp, (g * p0->mass * p1->mass) / r3, &tmp);
                        vector_add(&f, &f, &tmp);
                };
        };
        
        vector_scale(&f, 1 / p0->mass, &f);
        vector_copy(&p0->acceleration, &f);
};                        

static void update_positions(GList *particles_list, double h, double g) {
        PARTICLE *p;
        int i;

        /* For every particle, save initial values for position and velocity (for this step) */
        for(i = 0; i < g_list_length(particles_list); i++) {
                p = (PARTICLE *)g_list_nth_data(particles_list, i);
                vector_copy(&p->v_k0, &p->velocity);
                vector_copy(&p->p_k0, &p->position);                
        };
        
        /* For every particle, compute k1 */
        for(i = 0; i < g_list_length(particles_list); i++) {
                p = (PARTICLE *)g_list_nth_data(particles_list, i);
                update_acceleration(p, particles_list, g);
                vector_copy(&p->v_k1, &p->acceleration);
                vector_copy(&p->p_k1, &p->velocity);
        };
        
        /* For every particle, adjust position and velocity in order to compute k2 */
        for(i = 0; i < g_list_length(particles_list); i++) {
                p = (PARTICLE *)g_list_nth_data(particles_list, i);
                /* Vi = VK0i + (h / 2) * VK1i */
                vector_scale(&p->velocity, h / 2.0, &p->v_k1);
                vector_add(&p->velocity, &p->velocity, &p->v_k0);
                /* Pi = PK0i + (h / 2) * PK1i */
                vector_scale(&p->position, h / 2.0, &p->p_k1);
                vector_add(&p->position, &p->position, &p->p_k0);
        };
        
        /* For every particle, compute k2 */
        for(i = 0; i < g_list_length(particles_list); i++) {
                p = (PARTICLE *)g_list_nth_data(particles_list, i);
                update_acceleration(p, particles_list, g);
                vector_copy(&p->v_k2, &p->acceleration);
                vector_copy(&p->p_k2, &p->velocity);
        };
        
        /* For every particle, adjust position and velocity in order to compute k3 */
        for(i = 0; i < g_list_length(particles_list); i++) {
                p = (PARTICLE *)g_list_nth_data(particles_list, i);
                /* Vi = VK0i + (h / 2) * VK2i */
                vector_scale(&p->velocity, h / 2.0, &p->v_k2);
                vector_add(&p->velocity, &p->velocity, &p->v_k0);
                /* Pi = PK0i + (h / 2) * PK2i */
                vector_scale(&p->position, h / 2.0, &p->p_k2);
                vector_add(&p->position, &p->position, &p->p_k0);
        };
        
        /* For every particle, compute k3 */
        for(i = 0; i < g_list_length(particles_list); i++) {
                p = (PARTICLE *)g_list_nth_data(particles_list, i);
                update_acceleration(p, particles_list, g);
                vector_copy(&p->v_k3, &p->acceleration);
                vector_copy(&p->p_k3, &p->velocity);
        };
        
        /* For every particle, adjust position and velocity in order to compute k4 */
        for(i = 0; i < g_list_length(particles_list); i++) {
                p = (PARTICLE *)g_list_nth_data(particles_list, i);
                /* Vi = VK0i + h * VK3i */
                vector_scale(&p->velocity, h, &p->v_k3);
                vector_add(&p->velocity, &p->velocity, &p->v_k0);
                /* Pi = PK0i + h * PK3i */
                vector_scale(&p->position, h, &p->p_k3);
                vector_add(&p->position, &p->position, &p->p_k0);
        };
        
        /* For every particle, compute k4 */
        for(i = 0; i < g_list_length(particles_list); i++) {
                p = (PARTICLE *)g_list_nth_data(particles_list, i);
                update_acceleration(p, particles_list, g);
                vector_copy(&p->v_k4, &p->acceleration);
                vector_copy(&p->p_k4, &p->velocity);
        };        
        
        
        /* For every particle compute final position and velocity */
        for(i = 0; i < g_list_length(particles_list); i++) {
                p = (PARTICLE *)g_list_nth_data(particles_list, i);
                /* Vi = VK0i + (h / 6) * (VK1i + 2 * VK2i + 2 * VK3i + VK4i) */
                vector_scale(&p->v_k1, h / 6, &p->v_k1);
                vector_scale(&p->v_k2, h / 3, &p->v_k2);
                vector_scale(&p->v_k3, h / 3, &p->v_k3);
                vector_scale(&p->v_k4, h / 6, &p->v_k4);
                vector_copy(&p->velocity, &p->v_k0);
                vector_add(&p->velocity, &p->velocity, &p->v_k1);
                vector_add(&p->velocity, &p->velocity, &p->v_k2);
                vector_add(&p->velocity, &p->velocity, &p->v_k3);
                vector_add(&p->velocity, &p->velocity, &p->v_k4);
                
                /* Pi = PK0i + (h / 6) * (PK1i + 2 * PK2i + 2 * PK3i + PK4i) */
                vector_scale(&p->p_k1, h / 6, &p->p_k1);
                vector_scale(&p->p_k2, h / 3, &p->p_k2);
                vector_scale(&p->p_k3, h / 3, &p->p_k3);
                vector_scale(&p->p_k4, h / 6, &p->p_k4);
                vector_copy(&p->position, &p->p_k0);
                vector_add(&p->position, &p->position, &p->p_k1);
                vector_add(&p->position, &p->position, &p->p_k2);
                vector_add(&p->position, &p->position, &p->p_k3);
                vector_add(&p->position, &p->position, &p->p_k4);
        };               
};

static int check_step(GList *particles_list, double k) {
        PARTICLE *p;
        int i;
        
        /* Check if there is a particle that moved by a distance longer than its minimum distance */
        for(i = 0; i < g_list_length(particles_list); i++) {
                VECTOR tmp;
                double d;

                p = (PARTICLE *)g_list_nth_data(particles_list, i);
                vector_sub(&tmp, &p->position, &p->p0);
                d = vector_module(&tmp);
                
                if(d > (k * p->min_distance)) return FALSE;
        };
        
        return TRUE;
};

